/*  http://luftkvalitet.info is a site that issues warnings for air quality in Norway.
 *  This code is an IOT approach to get to these warnings, using a pushbutto, 5 through-hole
 *  5mm NeoPixels, and Adafruit ESP8266 Huzzah and a Li-Ion rechargable battery.
 *  
 *  Fetching a JSON document via https://api.nilu.no is handled by a local network 
 *  webserver, and published over HTTP to minimize code on part of the ESP8266 this runs on.
 *  
 *  The program fetches the document over WiFi and parses the JSON to get the current levels.
 *  The chosen station measures PM10, PM2.5 and NO2 and the JSON contains hex rgb color codes
 *  for the 4 different warning levels.
 *  These hex values are passed along to color mixing rgb LEDs to act as a user interface.
 *  
 *  Everything is mounted with a battery inside a small enclosure, with a button wired to
 *  the RST pin and GND, allowing for a reset to wake from deepsleep. This pushbutton is
 *  ultimately what is used for fetching and displaying the air quality status.
 *  author: Endre Mathistad, endre (at) variabel (dot) no
*/
#include <Adafruit_NeoPixel.h>
#ifdef __AVR__
  #include <avr/power.h>
#endif
#include <ESP8266WiFi.h>
#include <ESP8266WiFiMulti.h>
#include <ArduinoJson.h>
#include <ESP8266HTTPClient.h>
#include <string.h>
#define NUMLEDS 5
#define PIN 15
#define STATUS_CONNECTED 0    // WiFi indicator LED (as pixel #)
#define STATUS_DATA 1         // Data indicator LED
#define OK 1                  // More readable code
#define FAIL 0

#define MAX_CONN_TRIES 15     // Will deep sleep after max number
#define MAX_DATA_TRIES 5      // of failed attempts of either kind

const char* ssid     = "somewifi";
const char* password = "somepwd";
const char* httpRequest = "http://some.local.server/whatever.json";

// Simplify color selection for status LEDs
int LEDCOL[][3] = {
  {25, 0, 0},
  {0, 25, 0},
  {25, 25,0},
  {5, 10, 15},
  {0, 0, 0}
};
#define REDLED 0
#define GRNLED 1
#define YELLED 2
#define LOWLIT 3
#define ALLOFF 4

int conn_state = FAIL;        // State tracking
int data_state = FAIL;
int connectAttempt = 0;       // Just counters, limits are hard coded :P
int retrieveAttempt = 0;

Adafruit_NeoPixel strip = Adafruit_NeoPixel(NUMLEDS, PIN, NEO_RGB + NEO_KHZ400);

ESP8266WiFiMulti WiFiMulti;

void setup() {
  strip.begin();
  int i;
  for (i = 0;i < NUMLEDS; ++i) {
    lightMe(i, LOWLIT); // Initialize all pixels to 'LOWLIT'
  }
  delay(10);
  WiFiMulti.addAP(ssid, password);
}

void loop() {
  lightMe(STATUS_CONNECTED, YELLED);
  if((WiFiMulti.run() == WL_CONNECTED)) {
    conn_state = OK;
    blinkMe(50, STATUS_CONNECTED, GRNLED);
    lightMe(STATUS_DATA, YELLED);
    HTTPClient http;
    http.begin(httpRequest);
    int docSize;
    int httpCode = http.GET();
    if (httpCode == HTTP_CODE_OK && (docSize = http.getSize()) > 0) {
      blinkMe(50, STATUS_DATA, GRNLED);
      data_state = OK;

      // get tcp stream
      WiFiClient * stream = http.getStreamPtr();

      // prepare array for parsing
      char JSONDoc[1000];

      // read all data from server
      while(http.connected()) {
        blinkMe(50, STATUS_DATA, YELLED);
        stream->readBytes(JSONDoc, docSize);
      }
   
      // JSON Parsing
      StaticJsonBuffer<1000> JSONBuffer;
      JsonArray& root= JSONBuffer.parseArray(JSONDoc);

      // Failed to parse?
      if (!root.success()) {
        data_state = FAIL;
      } else {
        blinkMe(50, STATUS_DATA, GRNLED);
        int i;
        for (i = 0; i < 3; ++i) {
          // Make pointers to component and color value from JSON buffer
          const char* hexstring = root[i]["color"].asString();
          // Get hex number and convert to long, ref:
          //    http://stackoverflow.com/questions/23576827/arduino-convert-a-sting-hex-ffffff-into-3-int
          int r, g, b;
          int number = (int) strtol( &hexstring[0], NULL, 16);
          // And split into r, g, b values
          strip.setPixelColor(i+2, (r = number >> 16), (g = number >> 8 & 0xFF), (b = number & 0xFF));
          // We know that we have three entries, and also have 5 LEDs, of which
          // the last three will be used to display the values from JSON.
          // Hence i+2 is pin # of corresponding LED
        }
        strip.show();
      }
    } 
  }
  if (conn_state == FAIL) {
    blinkMe(350, STATUS_CONNECTED, REDLED);
    delay(500); // Failed to connect to WiFi
    ++connectAttempt;
    if (connectAttempt < MAX_CONN_TRIES) {
      return;
    } else {
      goToSleep();
    }
  } else {
    if ( data_state == FAIL) {
      blinkMe(350, STATUS_DATA, REDLED);
      delay(3000); // Failed to retrieve HTTP
      ++retrieveAttempt;
      if (retrieveAttempt < MAX_DATA_TRIES) {
        return;
      } else {
         goToSleep();
      }
    }
  }
  delay(2500);
  goToSleep();
}

void goToSleep() {
  int i;
  for (i = NUMLEDS; i >= 0; --i) {
    lightMe(i, ALLOFF);
  }
  ESP.deepSleep(0);
}

void lightMe(int led, int col) {
  int r = LEDCOL[col][0];
  int g = LEDCOL[col][1];
  int b = LEDCOL[col][2];
  strip.setPixelColor(led, r, g, b);
  strip.show();
}

void blinkMe(int onTime, int led, int col) {
  int r = LEDCOL[col][0];
  int g = LEDCOL[col][1];
  int b = LEDCOL[col][2];
  strip.setPixelColor(led, r, g, b);
  strip.show();
  delay(onTime);
  lightMe(led, LOWLIT);
  delay(50);
}
